#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.cause.errors import Codeable, MeshException, CompatibleException, NotFoundException, ValidationException, \
    TimeoutException, NoProviderException
from mesh.cause.status import MeshCode

__all__ = (
    "Codeable", "MeshException", "CompatibleException", "NotFoundException", "ValidationException", "TimeoutException",
    "NoProviderException", "MeshCode")


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import time
from typing import Any, Dict

import mesh.log as log
import mesh.telemetry as telemetry
import mesh.tool as tool
from mesh.cause import MeshCode, Codeable, MeshException
from mesh.context import Mesh
from mesh.macro import spi
from mesh.mpc.digest import Digest
from mesh.mpc.filter import Filter, Invoker, Invocation, PROVIDER
from mesh.prsim import Metadata
from mesh.prsim.context import RunMode


@spi(name="provider", pattern=PROVIDER, priority=0x7fffff00)
class ProviderFilter(Filter):

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not attachments:
            attachments = {}
        Mesh.context().decode(attachments)

        digest = Digest()
        try:
            ret = invoker.run(invocation)
            digest.write("P", MeshCode.SUCCESS.get_code())
            return ret
        except BaseException as e:
            if isinstance(e, Codeable):
                digest.write("P", e.get_code())
            else:
                digest.write("P", MeshCode.SYSTEM_ERROR.get_code())
            if isinstance(e, MeshException):
                log.error(f"{digest.trace_id},{Mesh.context().get_urn()},{e.get_message}")
            raise e

    @staticmethod
    def resolve_timestamp(v: str) -> int:
        if tool.optional(v):
            return int(time.time() * 1000)
        try:
            return int(v)
        except Exception as e:
            log.error("Parse timestamp failed. ", e)
            return int(time.time() * 1000)

    @staticmethod
    def resolve_run_mode(v: str) -> RunMode:
        if tool.optional(v):
            return RunMode.ROUTINE
        try:
            return RunMode.get_by_code(int(v))
        except Exception as e:
            log.error("Parse run mode failed.", e)
            return RunMode.ROUTINE


@spi(name="telemetryProvider", pattern=PROVIDER, priority=0x7ffffff0)
class TelemetryProviderFilter(Filter):
    """
    recover telemetry context via mesh context for tracing
    """

    def invoke(self, invoker: Invoker, invocation: Invocation) -> Any:
        attachments: Dict[str, str] = invocation.get_parameters().get_attachments()
        if not telemetry.if_rebuild_span():
            return invoker.run(invocation)
        trace_id = attachments.get(Metadata.MESH_TRACE_ID.key(), tool.new_trace_id())
        with telemetry.build_via_remote(attachments, invocation.get_urn().string()):
            current_span = telemetry.get_current_span()
            current_span.set_attribute('mesh-trace-id', trace_id)
            return invoker.run(invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

"""Performance test of the RingBuffer class."""

import argparse
import cProfile
import collections
import functools
import logging
import multiprocessing
import os
import pstats
import random
import struct
import time
import zlib

import mesh.boost.ringbuffer as ringbuffer

FLAGS = argparse.ArgumentParser()

FLAGS.add_argument('--debug', action='store_true')

FLAGS.add_argument('--duration-seconds', action='store',
                   type=int, required=True)

FLAGS.add_argument('--slots', action='store', type=int, required=True)

FLAGS.add_argument('--slot-bytes', action='store', type=int, required=True)

FLAGS.add_argument('--readers', action='store',
                   type=int, required=True)

FLAGS.add_argument('--reader-burn-cpu-milliseconds',
                   action='store', type=int, default=0)

FLAGS.add_argument('--writes-per-second', action='store',
                   type=int, required=True)

FLAGS.add_argument('--verify_writes', action='store_true',
                   dest='verify_writes')
FLAGS.add_argument('--no-verify_writes',
                   action='store_false', dest='verify_writes')
FLAGS.set_defaults(verify_writes=True)


def profile(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        profiler = cProfile.Profile()
        try:
            return profiler.runcall(func, *args, **kwargs)
        finally:
            stats = pstats.Stats(profiler)
            stats.strip_dirs()
            stats.sort_stats('tottime')
            stats.print_stats()

    return wrapper


def sleep_generator(duration_seconds, writes_per_second):
    start = time.time()
    end = start + duration_seconds
    target_duration = 1 / writes_per_second

    seen_durations = collections.deque(maxlen=10)

    while True:
        before = time.time()
        if before >= end:
            return

        yield

        after = time.time()
        last_duration = after - before
        # Keep an average duration to better approximate the processing
        # time and keep the sleep time stable. Otherwise the variability will
        # cause this method to oversleep.
        seen_durations.append(last_duration)

        avg_duration = sum(seen_durations) / len(seen_durations)
        next_delay = target_duration - avg_duration

        if next_delay > 0:
            time.sleep(next_delay)


# Using a memoryview prevents copying when random_data slices this value.
_CACHED_RANDOM_DATA = memoryview(100 * os.urandom(10 * 10 ** 6))


def get_random_data(num_bytes):
    """Generate random input data from cached randomness.
    We do this because os.urandom can be very slow and that's not what this
    code is trying to load test.
    """
    index = random.randint(0, len(_CACHED_RANDOM_DATA) - num_bytes)
    return _CACHED_RANDOM_DATA[index:index + num_bytes]


def get_crc32(data):
    return zlib.crc32(data) & 0xffffffff


def generate_verifiable_data(num_bytes):
    random_size = num_bytes - 4
    random_data = get_random_data(random_size)
    crc = get_crc32(random_data)

    result = bytearray(num_bytes)
    result[:random_size] = random_data
    struct.pack_into('>I', result, random_size, crc)

    return result


def verify_data(data):
    random_size = len(data) - 4
    random_data = data[:random_size]
    found_crc = get_crc32(random_data)
    (expected_crc,) = struct.unpack_from('>I', data, random_size)

    assert expected_crc == found_crc, 'Expected crc %r, found crc %r' % (
        expected_crc, found_crc)


class Timing:

    def __init__(self, now=time.time):
        self.now = now
        self.start = None
        self.end = None
        self.duration = None

    def __enter__(self):
        self.start = self.now()
        return self

    def __exit__(self, *args, **kwargs):
        self.end = self.now()
        self.duration = self.end - self.start
        return False


_PRINT_LOCK = multiprocessing.Lock()


def print_process_stats(process, flags, slots, elapsed):
    slots_per_second = slots / elapsed.duration
    mb_per_second = flags.slot_bytes * slots_per_second / 1e6
    delta = slots_per_second - flags.writes_per_second
    percent_wrong = 100 * delta / flags.writes_per_second

    message = """%(action)s
%(slots)d slots in %(duration)f seconds
%(bandwidth)f MBytes per second
%(rate)f slots/second, %(delta).1f%% relative to target
""" % {
        'action': process,
        'slots': slots,
        'duration': elapsed.duration,
        'rate': slots_per_second,
        'bandwidth': mb_per_second,
        'delta': percent_wrong,
    }

    with _PRINT_LOCK:
        print(message)


# @profile
def writer(flags, out_ring):
    print_every = flags.writes_per_second

    with Timing() as elapsed:
        it = sleep_generator(flags.duration_seconds, flags.writes_per_second)
        writes = 0
        for _ in it:
            if flags.verify_writes:
                data = generate_verifiable_data(flags.slot_bytes)
            else:
                data = get_random_data(flags.slot_bytes)

            try:
                out_ring.try_write(data)
            except ringbuffer.WaitingForReaderError:
                logging.error('Write %d is waiting for readers', writes)
                continue

            writes += 1
            if writes and writes % print_every == 0:
                logging.debug('Wrote %d slots so far', writes)

        out_ring.writer_done()

    logging.debug('Exiting writer')
    print_process_stats('Writer', flags, writes, elapsed)


def burn_cpu(milliseconds):
    start = now = time.time()
    end = start + milliseconds / 1000
    while True:
        now = time.time()
        if now >= end:
            break
        for i in range(100):
            random.random() ** 1 / 2


# @profile
def reader(flags, in_ring, reader):
    print_every = flags.writes_per_second
    read_duration = 1 / flags.writes_per_second
    reads = 0

    with Timing() as elapsed:
        while True:
            try:
                data = in_ring.blocking_read(reader)
            except ringbuffer.WriterFinishedError:
                break

            if flags.verify_writes:
                verify_data(data)

            if flags.reader_burn_cpu_milliseconds:
                burn_cpu(flags.reader_burn_cpu_milliseconds)

            reads += 1
            if reads % print_every == 0:
                logging.info('%r read %d slots so far', reader, reads)

    logging.debug('Exiting reader %r', reader)
    print_process_stats('Reader %d' % id(reader), flags, reads, elapsed)


def get_buffer(flags):
    return ringbuffer.RingBuffer(
        slot_bytes=flags.slot_bytes,
        slot_count=flags.slots)


def main():
    flags = FLAGS.parse_args()
    print('Starting performance test with flags: %r' % flags)

    if flags.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    ring = get_buffer(flags)
    ring.new_writer()

    processes = [
        multiprocessing.Process(
            target=writer,
            args=(flags, ring))
    ]
    for i in range(flags.readers):
        processes.append(
            multiprocessing.Process(
                target=reader,
                args=(flags, ring, ring.new_reader()))
        )

    for process in processes:
        process.start()

    for process in processes:
        process.join()
        assert process.exitcode == 0


if __name__ == '__main__':
    main()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.kinds.environ import Environ
from mesh.kinds.license import License
from mesh.macro import index, serializable


@serializable
class CommerceLicense:

    @index(0)
    def cipher(self) -> str:
        return ""

    @index(5)
    def explain(self) -> License:
        return License()


@serializable
class CommerceEnviron:

    @index(0)
    def cipher(self) -> str:
        return ""

    @index(5)
    def explain(self) -> Environ:
        return Environ()

    @index(10)
    def node_key(self) -> str:
        return ""



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import index, serializable


@serializable
class Lattice:

    @index(0)
    def zone(self) -> str:
        """zone"""
        return ''

    @index(5)
    def cluster(self) -> str:
        """cluster"""
        return ''

    @index(10)
    def cell(self) -> str:
        """cell"""
        return ''

    @index(15)
    def group(self) -> str:
        """group"""
        return ''

    @index(20)
    def address(self) -> str:
        """address"""
        return ''


class Environ:

    @index(0)
    def version(self) -> str:
        """Node certification version"""
        return ''

    @index(5)
    def node_id(self) -> str:
        """节点ID，所有节点按照标准固定分配一个全网唯一nodeId."""
        return ''

    @index(10)
    def inst_id(self) -> str:
        """每一个节点有一个初始的机构ID作为该节点的拥有者."""
        return ''

    @index(15)
    def inst_name(self) -> str:
        """一级机构名称."""
        return ''

    @index(20)
    def root_crt(self) -> str:
        """每一个节点内拥有一副证书用于通信，该证书可以被动态替换."""
        return ''

    @index(25)
    def root_key(self) -> str:
        """每一个节点内拥有一副证书用于通信，该证书可以被动态替换."""
        return ''

    @index(30)
    def node_crt(self) -> str:
        """节点许可证书"""
        return ''

    @index(35)
    def lattice(self) -> Lattice:
        """Mesh data center."""
        return Lattice()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Type, Any

from mesh.macro import Inspector, mpi, T


class MethodInspector(Inspector):

    def __init__(self, macro: mpi, kind: Type[T], method: Any):
        self.macro = macro
        self.kind = kind
        self.method = method

    def get_signature(self) -> str:
        return self.method.__name__

    def get_type(self) -> Type[T]:
        return self.kind

    def get_name(self) -> str:
        return self.method

    def get_annotation(self, kind: Type[T]) -> T:
        return self.method.__annotations__.get(kind)

    def get_return_type(self) -> Type[T]:
        return self.method.__annotations__.get("return")

    def invoke(self, obj: Any, args: [Any]) -> Any:
        return self.method(*args)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

# https://stackoverflow.com/questions/33533148/how-do-i-type-hint-a-method-with-the-type-of-the-enclosing-class

from typing import Generic, Dict, List, Type, Any

from mesh.cause import NotFoundException
from mesh.macro.ark import T, ark
from mesh.macro.spi import spi, SPI


class Instance(Generic[T]):
    values: Dict[type, Any] = {}

    def __init__(self, name: str, kind: type, prototype=False, pattern='', priority=0) -> None:
        self.name = name
        self.kind = kind
        self.prototype = prototype
        self.pattern = pattern
        self.priority = priority

    def __lt__(self, other):
        return self.priority > other.priority

    def get_if_absent(self) -> T:
        if self.prototype:
            return self.create()

        value = self.values.get(self.kind, None)

        if value is None:
            value = self.values[self.kind] = self.create()

        return value

    def create(self) -> T:
        return self.kind()


class ServiceLoader(Generic[T]):
    loaders: Dict[Type[T], "ServiceLoader[T]"] = {}
    resources = {}

    def __init__(self, kind: type) -> None:
        self.providers: Dict[str, List[Instance]] = dict()
        self.kind = kind
        self.first = SPI.get_macro(kind).name
        self.types = {}

    @staticmethod
    def load(kind: Type[T]) -> "ServiceLoader[T]":
        if not ServiceLoader.loaders.get(kind):
            ServiceLoader.loaders[kind] = ServiceLoader(kind)
        return ServiceLoader.loaders.get(kind)

    @staticmethod
    def resource(self, name: str) -> bytes:
        if hasattr(self.providers, name):
            return self.providers[name]
        return bytes("")

    def default_name(self) -> str:
        return self.first

    def get_default(self) -> T:
        return self.get(self.first)

    def get_default_cls(self) -> Type[T]:
        instance = self.get(self.first)
        return instance.__class__

    def get(self, name: str) -> T:
        instance = self.get_instance(name)
        if not instance:
            raise NotFoundException(f'No {self.kind.__name__} named {name} exist.')
        return instance.get_if_absent()

    def list(self, pattern: str) -> List[T]:
        instances: List[Instance] = []
        for _, iis in self.get_instances().items():
            for instance in iis:
                if '' == pattern or not pattern or instance.pattern == pattern:
                    instances.append(instance)
        instances = sorted(instances, key=lambda x: x.priority)
        services = []
        for instance in instances:
            services.append(instance.get_if_absent())
        return services

    def map(self) -> Dict[str, T]:
        instances = {}
        for name, iis in self.get_instances():
            for instance in iis:
                instances[name] = instance.get_if_absent()
        return instances

    def get_instance(self, name: str) -> Instance[T]:
        instances = self.get_instances().get(name, []) if name else self.get_instances().get(self.first, [])
        return instances[0] if instances.__len__() > 0 else None

    def get_instances(self) -> Dict[str, List[Instance[T]]]:
        if self.providers.__len__() > 0:
            return self.providers
        for spec in ark.providers(spi, self.kind):
            if not isinstance(spec.metadata, SPI):
                continue
            if spec.kind is self.kind:
                self.first = spec.metadata.name
                continue
            instance = Instance(spec.metadata.name, spec.kind, spec.metadata.prototype, spec.metadata.pattern,
                                spec.metadata.priority)
            instances = self.providers.get(spec.metadata.name, [])
            instances.append(instance)
            self.providers[spec.metadata.name] = instances

        return self.providers



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any

from mesh.macro import index, serializable


@serializable
class Outbound:

    @index(0)
    def code(self) -> str:
        return ''

    @index(1)
    def message(self) -> str:
        return ''

    @index(2)
    def cause(self) -> str:
        return ''

    @index(3)
    def content(self) -> Any:
        return None



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import mesh.log as log
import unittest

from mesh.macro import mps, index


class Property:
    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


@mps
class X:

    def __init__(self, getter) -> None:
        super().__init__()

    def getter(self):
        return self


@mps(name="x")
class Y:

    @index(0)
    def y(self) -> str:
        return '1'

    @index(5)
    def z(self) -> str:
        """
        Test z index.
        :return: z decorator value.
        """
        return '2'


class TestMacro(unittest.TestCase):

    def test_index(self):
        y = Y()
        self.invoke(y.z)
        self.invoke(y.y)
        y.z = '100'
        self.invoke(Y.z)
        self.invoke(y.z)

    @staticmethod
    def invoke(x: str):
        log.info(x)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any, Type, Dict

from mesh.cause import MeshCode
from mesh.codec import Codec, Json
from mesh.macro import mpi, ServiceLoader, Inspector, Parameters, Returns, Cause, T
from mesh.mpc.compiler import ATTACHMENTS, CODE, MESSAGE, CAUSE, CONTENT
from mesh.mpc.eden import Eden
from mesh.mpc.invoker import Invocation, Execution


class GenericParameters(Parameters, Dict):

    def map(self) -> Dict[str, Any]:
        return self

    def kind(self) -> type:
        return GenericParameters

    def get_arguments(self) -> [Any]:
        args = []
        for key, value in enumerate(self):
            if ATTACHMENTS == key:
                continue
            args.append(value)
        return args

    def set_arguments(self, arguments: [Any]):
        """ No implements """
        pass

    def get_attachments(self) -> Dict[str, str]:
        if not hasattr(self, ATTACHMENTS):
            return dict()
        attachments = self.get(ATTACHMENTS)
        if type(attachments) == dict:
            return attachments
        codec = ServiceLoader.load(Codec).get(Json)
        return codec.decode(attachments, Dict[str, str])

    def set_attachments(self, attachments: Dict[str, str]):
        self.__setitem__(ATTACHMENTS, attachments)


class GenericReturns(Returns, Dict):

    def get_code(self) -> str:
        code = self.get(CODE, None)
        if code:
            return code
        return MeshCode.UNKNOWN.get_code()

    def set_code(self, code: str):
        self.__setitem__(CODE, code)

    def get_message(self) -> str:
        message = self.get(MESSAGE, None)
        if message:
            return message
        return ""

    def set_message(self, message: str):
        self.__setitem__(MESSAGE, message)

    def get_cause(self) -> Cause:
        cause = self.get(CAUSE, None)
        if not cause:
            return cause
        if cause and type(cause == Cause):
            return cause
        codec = ServiceLoader.load(Codec).get(Json)
        return codec.decode(self[CAUSE], Cause)

    def set_cause(self, cause: Cause):
        self.__setitem__(CAUSE, cause)

    def get_content(self) -> Any:
        return self.get(CONTENT, None)

    def set_content(self, content: Any):
        self.__setitem__(CONTENT, content)


class GenericExecution(Execution):

    def __init__(self, refer: mpi, invocation: Invocation):
        self.refer = refer
        self.invocation = invocation

    def schema(self) -> T:
        eden = ServiceLoader.load(Eden).get_default()
        return eden.refer(self.refer, self.kind(), self.method())

    def inspect(self) -> Inspector:
        return self.invocation.get_inspector()

    def intype(self) -> Type[Parameters]:
        return GenericParameters

    def retype(self) -> Type[Returns]:
        return GenericReturns

    def inflect(self) -> Parameters:
        return GenericParameters()

    def reflect(self) -> Returns:
        return GenericReturns()

    def run(self, invocation: Invocation) -> Any:
        eden = ServiceLoader.load(Eden).get_default()
        execution = eden.refer(self.refer, self.kind(), self.method())
        return execution.run(invocation)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import inspect
import os
import random
import sys
from pathlib import Path
from typing import List, Any, Type, Dict

from mesh.environ import URI, Addrs, System
from mesh.macro import T, Compatible
from mesh.tool.snowflake import UUID


def init():
    """ init function """
    pass


__uuid = UUID(random.randint(0, 31), random.randint(0, 31))


def anyone(*args: str) -> str:
    if not args:
        return ""

    for arg in args:
        if arg and "" != arg:
            return arg

    return ""


def ternary(expr: bool, left: T, right: T) -> T:
    if expr:
        return left
    return right


def get_property(dft: str, name: List[str]) -> str:
    return System.environ().get_property(dft, name)


def get_ip() -> str:
    return System.environ().get_ip()


def get_hostname() -> str:
    return System.environ().get_hostname()


def get_mesh_address() -> Addrs:
    return System.environ().get_mesh_address()


def get_mesh_runtime() -> URI:
    return System.environ().get_mesh_runtime()


def get_mesh_name() -> str:
    return System.environ().get_mesh_name()


def get_mesh_mode() -> int:
    return System.environ().get_mesh_mode()


def get_mesh_direct() -> str:
    return System.environ().get_mesh_direct()


def get_max_channels() -> int:
    return System.environ().get_max_channels()


def get_min_channels() -> int:
    return System.environ().get_min_channels()


def required(*args: Any) -> bool:
    if not args:
        return False

    for arg in args:
        if arg is not None and "" != arg:
            continue
        return False

    return True


def optional(v: Any) -> bool:
    return not required(v)


def new_trace_id() -> str:
    return f"{System.environ().get_ip_hex()}{next_id()}"


def new_span_id(span_id: str, index: int) -> str:
    if optional(span_id):
        return "0"
    if span_id.__len__() > 255:
        return "0"
    return f"{span_id}.{index}"


def split(v: str, sep: str) -> [str]:
    if not v or "" == v:
        return []
    return v.split(sep)


def get_declared_methods(reference: Type[T]) -> Dict[Type[T], List[Any]]:
    kinds: List[Type[T]] = []
    for k, v in vars(reference).items():
        if inspect.isclass(v) and issubclass(reference, v):
            kinds.append(v)
    if inspect.isabstract(reference):
        kinds.append(reference)
    methods: Dict[Type[T], List[Any]] = {}
    for kind in kinds:
        methods[kind] = []
        for method in inspect.getmembers(kind, inspect.isfunction):
            methods[kind].append(method[1])
        for method in inspect.getmembers(kind, inspect.ismethod):
            methods[kind].append(method[1])
    return methods


def next_id() -> str:
    return __uuid.new_id()


def get_origin(tp):
    return Compatible.get_origin(tp)


def get_args(tp):
    return Compatible.get_args(tp)


def pwd() -> str:
    """
    Returns the name of the project root directory.
    :return: Project root directory name
    """

    # stack trace history related to the call of this function
    frame_stack: [inspect.FrameInfo] = inspect.stack()

    # get info about the module that has invoked this function
    # (index=0 is always this very module, index=1 is fine as long this function is not called by some other
    # function in this module)
    frame_info: inspect.FrameInfo = frame_stack[1]

    # if there are multiple calls in the stacktrace of this very module, we have to skip those and take the first
    # one which comes from another module
    if frame_info.filename == __file__:
        for frame in frame_stack:
            if frame.filename != __file__:
                frame_info = frame
                break

    # path of the module that has invoked this function
    caller_path: str = frame_info.filename

    # absolute path of the of the module that has invoked this function
    caller_absolute_path: str = os.path.abspath(caller_path)

    # get the top most directory path which contains the invoker module
    paths: [str] = [p for p in sys.path if p in caller_absolute_path]
    paths.sort(key=lambda p: len(p))
    caller_root_path: str = paths[0]

    if not os.path.isabs(caller_path):
        # file name of the invoker module (eg: "mymodule.py")
        caller_module_name: str = Path(caller_path).name

        # this piece represents a subpath in the project directory
        # (eg. if the root folder is "myproject" and this function has ben called from myproject/foo/bar/mymodule.py
        # this will be "foo/bar")
        project_related_folders: str = caller_path.replace(os.sep + caller_module_name, '')

        # fix root path by removing the undesired subpath
        caller_root_path = caller_root_path.replace(project_related_folders, '')

    dir_name: str = Path(caller_root_path).name

    return dir_name



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import time
import os
import unittest
import mesh.asm as asm
import mesh.log as log
from mesh.macro import mpi
from mesh.prsim import Network
import mesh.telemetry as telemetry


os.environ.setdefault("MESH_MODE", "")
os.environ.setdefault("MESH_ADDRESS", "")
os.environ.setdefault("MESH_NAME", "abc")
os.environ.setdefault("OPEN_TELEMETRY_ENDPOINT", "http://localhost:4317")


class TestMTelemetry(unittest.TestCase):

    @mpi
    def network(self) -> Network:
        """"""
        pass

    def test_get_environ(self):
        spn = telemetry.get_current_span()
        print(f'{spn}')
        with telemetry.tracer.start_as_current_span("asm.init"):
            asm.init()
            with telemetry.tracer.start_as_current_span("get_environ"):
                environ = self.network.get_environ()
                assert environ
                log.info(environ.node_id)
            log.info("asm.init done")
        log.info("test_get_environ done")


if __name__ == '__main__':
    # atta = {}
    # atta['mesh-telemetry-trace-id'] = 'e76c7c2efa4c33e794cc00e6d9a9aae8'
    # telemetry.build_via_remote(atta, 'test1')
    unittest.main()
    time.sleep(5)
    # curl http://10.99.56.76:3100/api/traces/011e8df9a141e7c4c952a15643c0c956
    # tempo service



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Dict

from mesh.macro import index, serializable


@serializable
class Script:
    EXPRESSION = "EXPRESSION"
    VALUE = "VALUE"
    SCRIPT = "SCRIPT"

    @index(0)
    def code(self) -> str:
        return ""

    @index(5)
    def name(self) -> str:
        return ""

    @index(10)
    def desc(self) -> str:
        return ""

    @index(15)
    def kind(self) -> str:
        return ""

    @index(20)
    def expr(self) -> str:
        return ""

    @index(25)
    def attachment(self) -> Dict[str, str]:
        return {}



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import ctypes
import gc
import logging
import multiprocessing
import queue
import threading
import time
import unittest

import mesh.boost.ringbuffer as ringbuffer


class SlotArrayTest(unittest.TestCase):

    def setUp(self):
        self.array = ringbuffer.SlotArray(slot_bytes=20, slot_count=10)

    def test_read_empty(self):
        for data in self.array:
            self.assertEqual(b'', data)

    def test_read_write(self):
        self.array[0] = b'hello'
        self.array[1] = b''
        self.array[5] = b'how are'
        self.array[9] = b'you doing today?'

        self.assertEqual(b'hello', self.array[0])
        self.assertEqual(b'', self.array[1])
        self.assertEqual(b'how are', self.array[5])
        self.assertEqual(b'you doing today?', self.array[9])

    def test_write_too_big(self):
        try:
            self.array[3] = b'asdfkljasdlfkajsflkjasdfasdfkljasdf'
            self.fail()
        except ringbuffer.DataTooLargeError:
            pass


class TestException(Exception):
    pass


class ReadersWriterLockTest(unittest.TestCase):

    def setUp(self):
        self.lock = ringbuffer.ReadersWriterLock()
        self.assert_unlocked()
        self.result_queues = {}

    def assert_unlocked(self):
        self.assertEqual(0, self.lock.readers.value)
        self.assertFalse(self.lock.writer.value)

    def assert_readers(self, count):
        self.assertEqual(count, self.lock.readers.value)
        self.assertFalse(self.lock.writer.value)

    def assert_writer(self):
        self.assertEqual(0, self.lock.readers.value)
        self.assertTrue(self.lock.writer.value)

    def reader_count(self):
        return self.lock.readers.value

    def async_run(self, func):
        def wrapper(result_queue):
            result = func()
            result_queue.put(result)

        result_queue = multiprocessing.Queue()
        process = multiprocessing.Process(
            target=wrapper,
            args=(result_queue,))

        self.result_queues[process] = result_queue

        process.start()
        return process

    def get_result(self, process):
        process.join()
        return self.result_queues[process].get()

    def test_read_then_write(self):
        with self.lock.for_read():
            self.assert_readers(1)

        self.assert_unlocked()

        with self.lock.for_write():
            self.assert_writer()

        self.assert_unlocked()

    def test_reentrant_readers(self):
        with self.lock.for_read():
            self.assert_readers(1)

            with self.lock.for_read():
                self.assert_readers(2)

                with self.lock.for_read():
                    self.assert_readers(3)

                self.assert_readers(2)

            self.assert_readers(1)

        self.assert_unlocked()

    def test_writer_blocks_reader(self):
        with self.lock.for_write():
            event = multiprocessing.Event()

            def test():
                self.assert_writer()

                # Caller will block until this event is released.
                event.set()

                with self.lock.for_read():
                    self.assert_readers(1)
                    return 'read'

            r = self.async_run(test)

            # Wait until we can confirm that the reader is locked out.
            event.wait()
            self.assert_writer()

        self.assertEqual('read', self.get_result(r))
        self.assert_unlocked()

    def test_writer_blocks_multiple_readers(self):
        with self.lock.for_write():
            before_read = multiprocessing.Barrier(3)
            during_read = multiprocessing.Barrier(2)
            after_read = multiprocessing.Barrier(2)

            def test():
                self.assert_writer()

                before_read.wait()

                with self.lock.for_read():
                    during_read.wait()
                    value = self.reader_count()
                    after_read.wait()
                    return value

            r1 = self.async_run(test)
            r2 = self.async_run(test)

            # Wait until we can confirm that all readers are locked out
            before_read.wait()
            self.assert_writer()

        self.assertEqual(2, self.get_result(r1))
        self.assertEqual(2, self.get_result(r2))
        self.assert_unlocked()

    def test_reader_blocks_writer(self):
        with self.lock.for_read():
            before_write = multiprocessing.Barrier(2)
            during_write = multiprocessing.Barrier(2)
            after_write = multiprocessing.Barrier(2)
            after_unlock = multiprocessing.Barrier(2)

            def test():
                self.assert_readers(1)

                before_write.wait()

                with self.lock.for_write():
                    self.assert_writer()
                    return 'written'

            writer = self.async_run(test)

            # Wait until we can confirm that all writers are locked out.
            before_write.wait()
            self.assert_readers(1)

        self.assertEqual('written', self.get_result(writer))
        self.assert_unlocked()

    def test_multiple_readers_block_writer(self):
        with self.lock.for_read():
            before_read = multiprocessing.Barrier(3)
            after_read = multiprocessing.Barrier(2)

            def test_reader():
                self.assert_readers(1)

                with self.lock.for_read():
                    before_read.wait()
                    value = self.reader_count()
                    after_read.wait()
                    return value

            def test_writer():
                before_read.wait()

                with self.lock.for_write():
                    self.assert_writer()
                    return 'written'

            reader = self.async_run(test_reader)
            writer = self.async_run(test_writer)

            # Wait for the write to be blocked by multiple readers.
            before_read.wait()
            self.assert_readers(2)
            after_read.wait()

        self.assertEqual(2, self.get_result(reader))
        self.assertEqual('written', self.get_result(writer))
        self.assert_unlocked()

    def test_multiple_writers_block_each_other(self):
        with self.lock.for_write():
            before_write = multiprocessing.Barrier(2)

            def test():
                before_write.wait()

                with self.lock.for_write():
                    self.assert_writer()
                    return 'written'

            writer = self.async_run(test)

            before_write.wait()
            self.assert_writer()

        self.assertEqual('written', self.get_result(writer))
        self.assert_unlocked()

    def test_wait_for_write(self):
        event = multiprocessing.Event()
        wait_count = 0

        with self.lock.for_read():
            def test():
                with self.lock.for_write():
                    self.assert_writer()
                    event.set()
                    return 'written'

            writer = self.async_run(test)

            while not event.is_set():
                self.assert_readers(1)
                wait_count += 1
                self.lock.wait_for_write()
                self.assert_readers(1)

        self.assertEqual('written', self.get_result(writer))
        self.assert_unlocked()
        self.assertLessEqual(wait_count, 2)

    def test_wait_for_write__writer_already_waiting_for_reader(self):
        event = multiprocessing.Event()

        with self.lock.for_read():
            def test():
                event.set()
                with self.lock.for_write():
                    self.assert_writer()
                    event.set()
                    return 'written'

            writer = self.async_run(test)

            event.wait()
            # Force a context switch so the writer is waiting
            time.sleep(0.1)

            self.lock.wait_for_write()
            self.assert_readers(1)

        self.assertEqual('written', self.get_result(writer))
        self.assert_unlocked()

    def test_wait_for_write_without_lock(self):
        self.assert_unlocked()
        self.assertRaises(
            ringbuffer.InternalLockingError,
            self.lock.wait_for_write)

    def test_unlock_readers_on_exception(self):
        try:
            with self.lock.for_read():
                self.assert_readers(1)
                raise TestException
        except TestException:
            self.assert_unlocked()
        else:
            self.fail()

    def test_unlock_writer_on_exception(self):
        try:
            with self.lock.for_write():
                self.assert_writer()
                raise TestException
        except TestException:
            self.assert_unlocked()
        else:
            self.fail()


class Expecter:

    def __init__(self, ring, pointer, testcase):
        self.ring = ring
        self.pointer = pointer
        self.testcase = testcase

    def expect_index(self, i):
        self.testcase.assertEqual(i, self.pointer.get().index)

    def write(self, data):
        self.ring.try_write(data)

    def write_memory_view(self, data):
        view = memoryview(data)
        self.ring.try_write(view)

    def write_ctype(self, data):
        data_type = ctypes.c_double * len(data)
        cdata = data_type()
        cdata[:] = data
        self.ring.try_write(cdata)

    def _get_read_func(self, blocking):
        if blocking:
            return self.ring.blocking_read
        else:
            return self.ring.try_read

    def expect_read(self, expected_data, blocking=False):
        read = self._get_read_func(blocking)
        data = read(self.pointer)
        self.testcase.assertEqual(expected_data, data, 'Data was: %r' % data)

    def expect_waiting_for_writer(self):
        # There's no blocking version of this because the WaitingForWriterError
        # is what's used to determine when to block on the condition variable.
        self.testcase.assertRaises(
            ringbuffer.WaitingForWriterError,
            self.ring.try_read,
            self.pointer)

    def expect_waiting_for_reader(self):
        self.testcase.assertRaises(
            ringbuffer.WaitingForReaderError,
            self.ring.try_write,
            b'should not work')

    def writer_done(self):
        self.ring.writer_done()

    def expect_writer_finished(self, blocking=False):
        read = self._get_read_func(blocking)
        self.testcase.assertRaises(
            ringbuffer.WriterFinishedError,
            read,
            self.pointer)

    def expect_already_closed(self):
        self.testcase.assertRaises(
            ringbuffer.AlreadyClosedError,
            self.ring.try_write,
            b'should not work')

    def force_reader_sync(self):
        self.ring.force_reader_sync()

    def expect_try_read_type(self, type_or_class):
        data = self.ring.try_read(self.pointer)
        self.testcase.assertTrue(isinstance(data, type_or_class))


class AsyncProxy:

    def __init__(self, expecter, in_queue, error_queue):
        self.expecter = expecter
        self.in_queue = in_queue
        self.error_queue = error_queue
        self.runner = None

    def run(self):
        while True:
            item = self.in_queue.get()
            try:
                if item == 'done':
                    logging.debug('Exiting %r', self.runner)
                    return

                name, args, kwargs = item
                logging.debug('Running %s(%r, %r)', name, args, kwargs)
                try:
                    result = getattr(self.expecter, name)(*args, **kwargs)
                except Exception as e:
                    logging.exception(
                        'Problem running %s(*%r, **%r)', name, args, kwargs)
                    self.error_queue.put(e)
            finally:
                self.in_queue.task_done()

    def shutdown(self):
        self.in_queue.put('done')

    def __getattr__(self, name):
        func = getattr(self.expecter, name)

        def proxy(*args, **kwargs):
            self.expecter.testcase.assertTrue(
                self.runner,
                'Must call start_proxies() before setting test expectations')

            # This queue is used to sequence operations between functions
            # that are running asynchronously (threads or processes).
            self.in_queue.put((name, args, kwargs))

            # If this test function is running in blocking mode, that means
            # the locking and sequencing is built into the semantics of the
            # function call itself. That means we can skip waiting for the
            # asynchronous function to consume the queue before letting
            # subsequent test methods run.
            if kwargs.get('blocking'):
                # Allow a context switch so the asynchronous function has
                # a chance to actually start the function call.
                time.sleep(0.1)
            else:
                self.in_queue.join()

        return proxy


class RingBufferTestBase:

    def setUp(self):
        self.ring = ringbuffer.RingBuffer(slot_bytes=100, slot_count=10)
        self.proxies = []
        self.error_queue = self.new_queue()

    def tearDown(self):
        for proxy in self.proxies:
            if proxy.runner:
                proxy.shutdown()
        for proxy in self.proxies:
            if proxy.runner:
                proxy.in_queue.join()
        if not self.error_queue.empty():
            raise self.error_queue.get()

        # Force child processes and pipes to be garbage collected, otherwise
        # we'll run out of file descriptors.
        gc.collect()

    def new_queue(self):
        raise NotImplementedError

    def run_proxy(self, proxy):
        raise NotImplementedError

    def start_proxies(self):
        for proxy in self.proxies:
            self.run_proxy(proxy)

    def new_reader(self):
        expecter = Expecter(self.ring, self.ring.new_reader(), self)
        proxy = AsyncProxy(expecter, self.new_queue(), self.error_queue)
        self.proxies.append(proxy)
        return proxy

    def new_writer(self):
        self.ring.new_writer()
        expecter = Expecter(self.ring, self.ring.writer, self)
        proxy = AsyncProxy(expecter, self.new_queue(), self.error_queue)
        self.proxies.append(proxy)
        return proxy

    def test_write_bytes(self):
        writer = self.new_writer()
        self.start_proxies()
        writer.write(b'this works')

    def test_write_string(self):
        writer = self.new_writer()
        self.start_proxies()
        self.assertTrue(self.error_queue.empty())
        writer.write('this does not work')
        error = self.error_queue.get()
        self.assertTrue(isinstance(error, TypeError))

    def test_write_bytearray(self):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        byte_list = [124, 129, 92, 3, 97]
        data = bytearray(byte_list)
        writer.write(data)

        expected_bytes = b'|\x81\\\x03a'
        self.assertListEqual(list(expected_bytes), byte_list)
        reader.expect_read(expected_bytes)

    def test_write_memoryview(self):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        data = b'|\x81\\\x03a'
        writer.write_memory_view(data)
        reader.expect_read(data)

    def test_write_ctype_array(self):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        data = [
            0.10547615602385774,
            0.7852261064650733,
            0.9641224591137485,
            0.7119325400788387,
            0.0351822948099656,
            0.7533559074003938,
            0.40285734175834087,
            0.9567564883196842,
            0.38539673218346415,
            0.2682555751644704,
        ]
        writer.write_ctype(data)

        expected_bytes = (
            b'\xe0X\xa1@|\x00\xbb?\xf3s\xe7\x7f\x92 \xe9?\xd8q\xe7W\x17\xda'
            b'\xee?)\x19\x13\xc0&\xc8\xe6?\x00\xcd6\xebi\x03\xa2?\x1f\x0f'
            b'\x11\xd9}\x1b\xe8?r\x8e\xf3(j\xc8\xd9?\x044r\xc8\xbf\x9d\xee?'
            b'\xe0\xa5-\x0eW\xaa\xd8?\xbcD\x93n\x19+\xd1?')
        reader.expect_read(expected_bytes)

        data_type = ctypes.c_double * len(data)
        expected = data_type.from_buffer_copy(expected_bytes)
        self.assertEqual(list(expected), data)

    def _do_read_single_write(self, blocking):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        writer.expect_index(0)
        writer.write(b'first write')
        writer.expect_index(1)

        reader.expect_index(0)
        reader.expect_read(b'first write', blocking=blocking)
        reader.expect_index(1)

    def test_read_is_bytes(self):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        writer.write(b'this works')
        reader.expect_try_read_type(bytearray)

    def test_read_single_write_blocking(self):
        self._do_read_single_write(True)

    def test_read_single_write_non_blocking(self):
        self._do_read_single_write(False)

    def _do_read_ahead_of_writes(self, blocking):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        reader.expect_waiting_for_writer()
        writer.write(b'first write')
        reader.expect_read(b'first write', blocking=blocking)

    def test_read_ahead_of_writes_blocking(self):
        self._do_read_ahead_of_writes(True)

    def test_read_ahead_of_writes_non_blocking(self):
        self._do_read_ahead_of_writes(False)

    def _do_two_reads_one_behind_one_ahead(self, blocking):
        r1 = self.new_reader()
        r2 = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        writer.write(b'first write')

        r1.expect_read(b'first write', blocking=blocking)
        r1.expect_waiting_for_writer()

        r2.expect_read(b'first write', blocking=blocking)
        r2.expect_waiting_for_writer()

    def test_two_reads_one_behind_one_ahead_blocking(self):
        self._do_two_reads_one_behind_one_ahead(True)

    def test_two_reads_one_behind_one_ahead_non_blocking(self):
        self._do_two_reads_one_behind_one_ahead(False)

    def test_write_conflict_first_slot(self):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        for i in range(self.ring.slot_count):
            writer.write(b'write %d' % i)

        # The writer has wrapped around and is now waiting for the reader
        # to free up a slot. They have the same index, but are different
        # generations.
        reader.expect_index(0)
        writer.expect_index(0)
        writer.expect_waiting_for_reader()

        reader.expect_read(b'write 0')
        writer.write(b'now it works')

        for i in range(1, self.ring.slot_count):
            reader.expect_read(b'write %d' % i)

        reader.expect_index(0)
        reader.expect_read(b'now it works')

    def test_write_conflict_last_slot(self):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        last_slot = self.ring.slot_count - 1
        self.assertGreater(last_slot, 0)

        for i in range(last_slot):
            data = b'write %d' % i
            writer.write(data)
            reader.expect_read(data)

        writer.expect_index(last_slot)
        reader.expect_index(last_slot)

        # The reader's pointed at the last slot, now wrap around the writer
        # to catch up. They'll have the same index, but different generation
        # numbers.
        for i in range(self.ring.slot_count):
            data = b'write %d' % (self.ring.slot_count + i)
            writer.write(data)

        reader.expect_index(last_slot)
        writer.expect_index(last_slot)
        writer.expect_waiting_for_reader()

        reader.expect_read(b'write 10')
        writer.write(b'now it works')
        writer.expect_index(0)
        reader.expect_index(0)

    def test_create_reader_after_writing(self):
        writer = self.new_writer()
        self.start_proxies()

        self.new_reader()  # No error because no writes happened yet.

        writer.write(b'hello')
        self.assertRaises(
            ringbuffer.MustCreatedReadersBeforeWritingError,
            self.new_reader)

    def _do_read_after_close_beginning(self, blocking):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        writer.writer_done()
        reader.expect_writer_finished(blocking=blocking)

    def test_read_after_close_beginning_blocking(self):
        self._do_read_after_close_beginning(True)

    def test_read_after_close_beginning_non_blocking(self):
        self._do_read_after_close_beginning(False)

    def _do_close_before_read(self, blocking):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        writer.write(b'fill the buffer')
        writer.writer_done()
        writer.expect_index(1)

        reader.expect_read(b'fill the buffer')
        reader.expect_writer_finished(blocking=blocking)
        reader.expect_index(1)

    def test_close_before_read_blocking(self):
        self._do_close_before_read(True)

    def test_close_before_read_non_blocking(self):
        self._do_close_before_read(False)

    def _do_close_after_read(self, blocking):
        reader = self.new_reader()
        writer = self.new_writer()
        self.start_proxies()

        writer.write(b'fill the buffer')

        reader.expect_read(b'fill the buffer')
        reader.expect_waiting_for_writer()
        reader.expect_index(1)

        writer.writer_done()
        writer.expect_index(1)

        reader.expect_writer_finished(blocking=blocking)

    def test_close_after_read_blocking(self):
        self._do_close_after_read(True)

    def test_close_after_read_non_blocking(self):
        self._do_close_after_read(False)

    def test_close_then_write(self):
        writer = self.new_writer()
        self.start_proxies()

        writer.write(b'one')
        writer.writer_done()
        writer.expect_already_closed()

    def test_blocking_readers_wake_up_after_write(self):
        writer = self.new_writer()
        r1 = self.new_reader()
        r2 = self.new_reader()
        self.start_proxies()

        r1.expect_read(b'write after read', blocking=True)
        r2.expect_read(b'write after read', blocking=True)

        writer.write(b'write after read')

    def test_blocking_readers_wake_up_after_close(self):
        writer = self.new_writer()
        r1 = self.new_reader()
        r2 = self.new_reader()
        self.start_proxies()

        r1.expect_writer_finished(blocking=True)
        r2.expect_writer_finished(blocking=True)

        writer.writer_done()

    def test_force_reader_sync(self):
        writer = self.new_writer()
        r1 = self.new_reader()
        r2 = self.new_reader()
        self.start_proxies()

        writer.write(b'one')
        writer.write(b'two')
        writer.write(b'three')

        writer.expect_index(3)
        r1.expect_index(0)
        r2.expect_index(0)

        writer.force_reader_sync()
        r1.expect_index(3)
        r2.expect_index(3)

    def _do_multiple_writers(self, blocking):
        w1 = self.new_writer()
        w2 = self.new_writer()
        reader = self.new_reader()
        self.start_proxies()

        w1.write(b'aaa')
        w1.expect_index(1)
        w2.expect_index(1)

        w2.write(b'bbb')
        w1.expect_index(2)
        w2.expect_index(2)

        w2.write(b'ccc')
        w1.expect_index(3)
        w2.expect_index(3)

        w1.write(b'ddd')
        w1.expect_index(4)
        w2.expect_index(4)

        reader.expect_read(b'aaa', blocking=blocking)
        reader.expect_read(b'bbb', blocking=blocking)
        reader.expect_read(b'ccc', blocking=blocking)
        reader.expect_read(b'ddd', blocking=blocking)

    def test_multiple_writers_blocking(self):
        self._do_multiple_writers(True)

    def test_multiple_writers_non_blocking(self):
        self._do_multiple_writers(False)

    def _do_test_multiple_writers_close(self, blocking):
        w1 = self.new_writer()
        w2 = self.new_writer()
        reader = self.new_reader()
        self.start_proxies()

        w1.write(b'aaa')
        w1.writer_done()

        w2.write(b'bbb')
        w2.writer_done()

        reader.expect_read(b'aaa', blocking=blocking)
        reader.expect_read(b'bbb', blocking=blocking)
        reader.expect_writer_finished(blocking=blocking)

    def test_multiple_writers_close_blocking(self):
        self._do_test_multiple_writers_close(True)

    def test_multiple_writers_close_non_blocking(self):
        self._do_test_multiple_writers_close(False)

    def _do_start_read_before_writer_setup(self, blocking):
        reader = self.new_reader()
        self.start_proxies()
        reader.expect_writer_finished(blocking=blocking)

    def test_start_read_before_writer_setup_blocking(self):
        self._do_start_read_before_writer_setup(True)

    def test_start_read_before_writer_setup_non_blocking(self):
        self._do_start_read_before_writer_setup(False)


class ThreadingTest(RingBufferTestBase, unittest.TestCase):

    def new_queue(self):
        return queue.Queue()

    def run_proxy(self, proxy):
        thread = threading.Thread(target=proxy.run)
        proxy.runner = thread
        thread.daemon = True
        thread.start()


class MultiprocessingTest(RingBufferTestBase, unittest.TestCase):

    def new_queue(self):
        return multiprocessing.JoinableQueue()

    def run_proxy(self, proxy):
        process = multiprocessing.Process(target=proxy.run)
        proxy.runner = process
        process.daemon = True
        process.start()


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from abc import abstractmethod


class CauseHandler:

    @abstractmethod
    def handle(self):
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import atexit
import signal
import sys
import traceback
from typing import Any, Type

import mesh.log as log


class Runtime(object):

    def __init__(self):
        self.sys_exist_hook = sys.exit
        self.sys_except_hook = sys.excepthook
        sys.exit = self.exit_hook
        sys.excepthook = self.except_hook

    def exit_hook(self, status=0, *args):
        if self.sys_exist_hook:
            self.sys_exist_hook(status)

    def except_hook(self, kind: Type[BaseException], e: BaseException, trace_type: Any, *args):
        if self.sys_except_hook:
            self.sys_except_hook(kind, e, trace_type)

    @staticmethod
    def safe_exec_hook(hook: Any):
        try:
            hook()
        except BaseException as e:
            log.error(f"{traceback.format_exception_only(type(e), e)}")

    def add_shutdown_hook(self, hook: Any):
        def safe_hook():
            return self.safe_exec_hook(hook)

        atexit.register(safe_hook)
        signal.signal(signal.SIGTERM, safe_hook)
        signal.signal(signal.SIGINT, safe_hook)

    @staticmethod
    def get_runtime() -> "Runtime":
        return runtime


runtime = Runtime()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Any, Dict, Type

from mesh.macro.cause import Cause
from mesh.macro.types import T


class Returns(ABC):

    @abstractmethod
    def get_code(self) -> str:
        """ Return code. """
        pass

    @abstractmethod
    def set_code(self, code: str):
        """ Return code. """
        pass

    @abstractmethod
    def get_message(self) -> str:
        """ Return message. """
        pass

    @abstractmethod
    def set_message(self, message: str):
        """ Return message. """
        pass

    @abstractmethod
    def get_cause(self) -> Cause:
        """ Return cause. """
        pass

    @abstractmethod
    def set_cause(self, cause: Cause):
        """ Return cause. """
        pass

    @abstractmethod
    def get_content(self) -> Any:
        """ Return content. """
        pass

    @abstractmethod
    def set_content(self, content: Any):
        """ Return content. """
        pass


class Parameters(ABC):

    @abstractmethod
    def map(self) -> Dict[str, Any]:
        """ Convert parameters to map. """
        pass

    @abstractmethod
    def kind(self) -> type:
        """ Parameters declared type. """
        pass

    @abstractmethod
    def get_arguments(self) -> [Any]:
        """ Generic arguments array. """
        pass

    @abstractmethod
    def set_arguments(self, *arguments):
        """ Generic arguments array. """
        pass

    @abstractmethod
    def get_attachments(self) -> Dict[str, str]:
        """ Get the generic attachments. The attributes will be serialized. The attachments are mutable. """
        pass

    @abstractmethod
    def set_attachments(self, attachments: Dict[str, str]):
        """ Attachment arguments. """
        pass


class Inspector(ABC):

    @abstractmethod
    def get_signature(self) -> str:
        """ Get the inspector signature. """
        pass

    @abstractmethod
    def get_type(self) -> Type[T]:
        """ Get the declared type of inspect object declared. """
        pass

    @abstractmethod
    def get_name(self) -> str:
        """ Get the name of inspector. """
        pass

    @abstractmethod
    def get_annotation(self, kind: Type[T]) -> T:
        """ Get the annotations of inspector. """
        pass

    @abstractmethod
    def get_return_type(self) -> Type[T]:
        """ Get the return type of inspector. """
        pass

    @abstractmethod
    def invoke(self, obj: Any, args: [Any]) -> Any:
        """ Invoke the inspector object. """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, Generic

from mesh.macro.ark import ark, T


class MPS(Generic[T]):

    def __init__(self, fn: T = None, name='', version='', proto='grpc', codec='json', flags=0, timeout=10000):
        """
        Metadata annotation for Serial Peripheral Interface. Can be used with {@link ServiceLoader#load(Class)}
        or dependency injection at compile time and runtime time.
        """
        self.urn = ""
        self.namespace = ""
        self.name = name
        self.version = version
        self.proto = proto
        self.codec = codec
        self.flags = flags
        self.timeout = timeout
        self.retries = 10
        self.node = ""
        self.inst = ""
        self.zone = ""
        self.cluster = ""
        self.cell = ""
        self.group = ""
        self.address = ""
        self.kind = ""
        self.lang = ""
        self.attrs = {}
        if name != '':
            self.name = name
            return
        if fn is None:
            return
        if type(fn) is str:
            self.name = fn
            return
        self.name = fn.__name__
        self.kind = fn
        ark.register(mps, self.name, self.kind, self)

    def __call__(self, *args, **kwargs) -> T:
        if not hasattr(self, 'kind') or self.kind == '':
            self.kind = args[0]

        ark.register(mps, self.name, self.kind, self)
        return self.kind

    @staticmethod
    def get_mps_if_present(target: Any) -> "MPS":
        if hasattr(target, '__mps__'):
            return target.__mps__

        return MPS()


def mps(fn: T = None, *, name='', version='', proto='grpc', codec='json', flags=0, timeout=10000) -> T:
    """
    Multi Provider Service. Mesh provider service.
    :param fn:
    :param name:
    :param version:
    :param proto:
    :param codec:
    :param flags: Service flag 1 asyncable 2 encrypt 4 communal.
    :param timeout:
    :return:
    """
    return MPS(fn, name, version, proto, codec, flags, timeout)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

# import socket
# import threading
# import time
#
# from prometheus_client.registry import REGISTRY, CollectorRegistry
#
# from mesh.kinds import Topic
# from mesh.metrics.collector import CPU_COLLECTOR, MEM_COLLECTOR, DISK_COLLECTOR, NETWORK_COLLECTOR
#
#
# # start_collector: start prom collector thread to gather system info
# def start_collector():
#     COLLECTOR_SCHEDULER.start()
#
#
# class CollectorScheduler(threading.Thread):
#
#     def __init__(self, thread_name='thread-prom-collector', interval=10):
#         threading.Thread.__init__(self, name=thread_name)
#         self.hostname = socket.gethostname()
#         if interval < 1:
#             self.interval = 10
#         else:
#             self.interval = interval
#         topic = Topic()
#         topic.topic = "mesh.prom.sample.export"
#         topic.code = "*"
#         self.bindings = topic
#         # TODO load publisher
#         self.publisher = {}
#         self.default_registry = REGISTRY
#         self.registry = CollectorRegistry(auto_describe=False)
#         self.registry.register(CPU_COLLECTOR)
#         self.registry.register(MEM_COLLECTOR)
#         self.registry.register(DISK_COLLECTOR)
#         self.registry.register(NETWORK_COLLECTOR)
#
#     def run(self):
#         while True:
#             self.collect()
#             # collect per 10s
#             time.sleep(self.interval)
#
#     def collect(self):
#         samples = []
#         self_metrics = self.registry.collect()
#         if self_metrics:
#             for metric in iter(self_metrics):
#                 print(metric)
#                 samples.append(metric)
#
#         metrics = self.default_registry.collect()
#         for metric in metrics:
#             print(metric)
#             samples.append(metric)
#
#         if len(samples) > 0:
#             print(samples)
#             event = {
#                 "metricFamilySamples": samples,
#                 "instanceId": self.hostname,
#                 "jobName": "pushgateway",
#             }
#             try:
#                 # TODO publish event
#                 # self.publish(event)
#                 print(event)
#             except:
#                 print("publisher publish Exception:")
#
#     def publish(self, payload):
#         # TODO active publish
#         # self.publisher.publish_with_topic(self.bindings, payload)
#         pass
#
#
# COLLECTOR_SCHEDULER = CollectorScheduler()
#
# # for local test
# if __name__ == "__main__":
#     scheduler = CollectorScheduler()
#     scheduler.run()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import Any, Sequence

from mesh.macro import index, serializable


@serializable
class Inbound:

    @index(0)
    def arguments(self) -> Sequence[Any]:
        return []

    @index(1)
    def attachments(self) -> {}:
        return {}



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from typing import Any, List, Generic

from mesh.macro.ark import T, ark


class Binding(Generic[T]):

    def __init__(self, fn: T = None, topic='', code='', version='', proto='grpc', codec='json', flags=0, timeout=10000,
                 meshable=True):
        """
        Metadata annotation for Serial Peripheral Interface. Can be used with {@link ServiceLoader#load(Class)}
        or dependency injection at compile time and runtime time.
        """
        self.topic = topic
        self.code = code
        self.version = version
        self.proto = proto
        self.codec = codec
        self.flags = flags
        self.timeout = timeout
        self.meshable = meshable
        if topic != '':
            self.topic = topic
            return
        if fn is None:
            return
        if type(fn) is str:
            self.name = fn
            return
        self.topic = str(type(fn))
        self.kind = fn

    def __call__(self, *args, **kwargs) -> T:
        if not hasattr(self, 'kind'):
            self.kind = args[0]

        if hasattr(self.kind, '__bindings__'):
            self.kind.__bindings__.append(self)
        else:
            self.kind.__bindings__ = [self]
        ark.register(binding, f"{self.topic}.{self.code}", self.kind, self)
        return self.kind

    @staticmethod
    def get_binding_if_present(target: Any) -> List["Binding"]:
        if hasattr(target.__class__, '__bindings__'):
            return target.__class__.__bindings__

        return []

    def matches(self, bindings: List["Binding"]) -> bool:
        if not bindings:
            return False
        for b in bindings:
            if b.topic == self.topic and b.code == self.code:
                return True
        return False


def binding(fn: T = None, *, topic='', code='', version='', proto='grpc', codec='json', flags=0, timeout=10000,
            meshable=True) -> T:
    """
    Multi Provider Service. Mesh provider service.
    :param fn:
    :param topic:
    :param code:
    :param version:
    :param proto:
    :param codec:
    :param flags:
    :param timeout:
    :param meshable:
    :return:
    """
    return Binding(fn, topic, code, version, proto, codec, flags, timeout, meshable)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import mesh.log as log
import types
import unittest
from abc import ABC, abstractmethod
from typing import Any

from mesh.macro import Proxy, InvocationHandler


class Foo(ABC):

    @abstractmethod
    def bar(self) -> str:
        pass


class ProxyHandler(InvocationHandler):

    def invoke(self, proxy: Any, func: types.FunctionType, *args, **kwargs):
        return "bar"


class TestProxy(unittest.TestCase):

    def test_get_environ(self):
        foo = Proxy.new_proxy(Foo, ProxyHandler())
        bar = foo.bar()
        log.info(bar)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import os
import time
import unittest

import mesh.asm as asm
import mesh.log as log
from mesh.macro import mpi
from mesh.prsim import Network
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
)
from opentelemetry.sdk.resources import SERVICE_NAME, Resource

resource = Resource(attributes={SERVICE_NAME: "gaia-mpi"})

# os.environ.setdefault("MESH_MODE", "4")
os.environ.setdefault("MESH_ADDRESS", "10.99.56.33")
os.environ.setdefault("MESH_NAME", "pandoroo")

provider = TracerProvider(resource=resource)
processor = BatchSpanProcessor(ConsoleSpanExporter())
span_processor = BatchSpanProcessor(OTLPSpanExporter(endpoint="http://10.99.56.76:4318/v1/traces"))
provider.add_span_processor(processor)
provider.add_span_processor(span_processor)
trace.set_tracer_provider(provider)

tracer = trace.get_tracer(os.environ.get("MESH_NAME"))


class TestGrpc(unittest.TestCase):

    @mpi
    def network(self) -> Network:
        """"""
        pass

    def test_get_environ(self):
        with tracer.start_as_current_span("asm.init"):
            asm.init()
            with tracer.start_as_current_span("get_environ"):
                environ = self.network.get_environ()
                assert environ
                log.info(environ.node_id)
            log.info("asm.init done")
        log.info("test_get_environ done")


if __name__ == '__main__':
    unittest.main()
    time.sleep(15)
    # curl http://10.99.56.76:3100/api/traces/011e8df9a141e7c4c952a15643c0c956
    # tempo service



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import abstractmethod, ABC
from asyncio import futures
from typing import Any, Dict
from typing import Type, Generic

from mesh.macro import T, Inspector, Parameters, Returns
from mesh.macro import spi
from mesh.mpc.urn import URN


@spi(name='invoker', pattern="*")
class Invoker(Generic[T], ABC):
    """
    Invoke the next invoker.
    """

    @abstractmethod
    def run(self, invocation: "Invocation") -> Any:
        pass


class Invocation(ABC):

    @abstractmethod
    def get_proxy(self) -> Invoker[T]:
        """
         Get the delegate target object.
        :return: proxy target
        """
        pass

    @abstractmethod
    def get_inspector(self) -> Inspector:
        """
        Get the invocation inspector.
        :return: inspector
        """
        pass

    @abstractmethod
    def get_parameters(self) -> Parameters:
        """
        Invoke parameters. include arguments and parameters.
        :return: parameters
        """
        pass

    @abstractmethod
    def get_arguments(self) -> [Any]:
        """
        Invoke parameters.
        :return: arguments
        """
        pass

    @abstractmethod
    def get_attachments(self) -> Dict[str, str]:
        """
        Get the attachments. The attributes will be serialized.
        :return: attachments
        """
        pass

    @abstractmethod
    def get_execution(self) -> "Execution":
        """ Get the invocation execution. """
        pass

    @abstractmethod
    def is_futures(self) -> bool:
        """
        Is the method return future.
        :return:
        """
        pass

    @abstractmethod
    def get_urn(self) -> URN:
        """
        Get the invoke urn.
        :return:
        """
        pass


class Execution(Invoker[T], Generic[T], ABC):
    """

    """

    @abstractmethod
    def schema(self) -> T:
        """ Execution schema. """
        pass

    @abstractmethod
    def inspect(self) -> Inspector:
        """ Inspect execution. """
        pass

    @abstractmethod
    def intype(self) -> Type[Parameters]:
        """ Execution input type. """
        pass

    @abstractmethod
    def retype(self) -> Type[Returns]:
        """ Execution output return type. """
        pass

    @abstractmethod
    def inflect(self) -> Parameters:
        """ Reflect input type. """
        pass

    @abstractmethod
    def reflect(self) -> Returns:
        """ Reflect output return type. """
        pass


class ServiceInvocation(Generic[T], Invocation):

    def __init__(self, proxy: Invoker[T], inspector: Inspector, parameters: Parameters, execution: Execution, urn: URN):
        self.proxy = proxy
        self.inspector = inspector
        self.parameters = parameters
        self.execution = execution
        self.urn = urn

    def get_proxy(self) -> Invoker[T]:
        return self.proxy

    def get_inspector(self) -> Inspector:
        return self.inspector

    def get_parameters(self) -> Parameters:
        return self.parameters

    def get_arguments(self) -> [Any]:
        return self.parameters.get_arguments()

    def get_attachments(self) -> Dict[str, str]:
        return self.parameters.get_attachments()

    def get_execution(self) -> Execution:
        return self.execution

    def is_futures(self) -> bool:
        rt = self.inspector.get_return_type()
        if futures.isfuture(rt):
            return True
        return False

    def get_urn(self) -> URN:
        return self.urn
